import logging
import re
from unittest.mock import MagicMock, PropertyMock, call, patch

import iris
import numpy as np
import pytest
from conftest import create_sftlf_cube

from netcdf_scm.iris_cube_wrappers import SCMCube
from netcdf_scm.weights import (
    DEFAULT_REGIONS,
    WEIGHTS_FUNCTIONS_WITHOUT_AREA_WEIGHTING,
    CubeWeightCalculator,
    get_default_sftlf_cube,
    get_land_weights,
    get_ocean_weights,
    get_weights_for_area,
)


@patch("netcdf_scm.weights.get_land_weights")
@patch("netcdf_scm.weights.get_nh_weights")
def test_get_scm_masks(mock_nh_weights, mock_land_weights, test_all_cubes):
    tsftlf_cube = "mocked 124"

    land_weights = np.array([[1, 0, 0, 1], [1, 0, 1, 0], [1, 1, 0, 1]])
    mock_land_weights.return_value = land_weights

    nh_weights = np.array([[1, 1, 1, 1], [1, 1, 1, 1], [0, 0, 0, 0]])
    mock_nh_weights.return_value = nh_weights

    nh_land_weights = np.array([[1, 0, 0, 1], [1, 0, 1, 0], [0, 0, 0, 0]])
    # check our logic while we're here
    np.testing.assert_array_equal(nh_weights * land_weights, nh_land_weights)

    area_weights = test_all_cubes.get_area_weights()
    expected = {
        k: area_weights * v
        for k, v in {
            "World": np.full(nh_weights.shape, 1),
            "World|Northern Hemisphere|Land": nh_land_weights,
            "World|Southern Hemisphere|Land": (1 - nh_weights) * land_weights,
            "World|Northern Hemisphere|Ocean": nh_weights * (100 - land_weights),
            "World|Southern Hemisphere|Ocean": (1 - nh_weights) * (100 - land_weights),
            "World|Land": land_weights,
            "World|Ocean": 100 - land_weights,
            "World|Northern Hemisphere": nh_weights,
            "World|Southern Hemisphere": 1 - nh_weights,
        }.items()
    }

    with patch.dict(
        WEIGHTS_FUNCTIONS_WITHOUT_AREA_WEIGHTING,
        {"World|Northern Hemisphere": mock_nh_weights, "World|Land": mock_land_weights},
    ):
        weighter = CubeWeightCalculator(test_all_cubes, sftlf_cube=tsftlf_cube)
        result = weighter.get_weights(DEFAULT_REGIONS)

    for label, array in expected.items():
        np.testing.assert_array_equal(array, result[label])
    mock_land_weights.assert_called_with(
        weighter, test_all_cubes, sftlf_cube=tsftlf_cube
    )
    mock_nh_weights.assert_called_with(weighter, test_all_cubes, sftlf_cube=tsftlf_cube)


@pytest.mark.parametrize("with_bounds", [True, False])
@patch("netcdf_scm.weights.get_nh_weights")
def test_get_scm_masks_no_land_available(
    mock_nh_weights, with_bounds, test_all_cubes, caplog
):
    caplog.set_level(logging.WARNING, logger="netcdf_scm.iris_cube_wrappers")
    test_all_cubes.get_metadata_cube = MagicMock(side_effect=OSError)

    nh_weights = np.array([[1, 1, 1, 1], [1, 1, 1, 1], [0, 0, 0, 0]])
    mock_nh_weights.return_value = nh_weights
    default_sftlf_cube = get_default_sftlf_cube()
    default_sftlf_cube = default_sftlf_cube.regrid(
        # AreaWeighted() in future but too slow now
        test_all_cubes.cube,
        iris.analysis.Linear(),
    )
    expected_land_weights = default_sftlf_cube.data

    area_weights = test_all_cubes.get_area_weights()
    expected = {
        k: area_weights * v
        for k, v in {
            "World": np.full(nh_weights.shape, 1),
            "World|Northern Hemisphere": nh_weights,
            "World|Northern Hemisphere|Land": nh_weights * expected_land_weights,
            "World|Southern Hemisphere": 1 - nh_weights,
        }.items()
    }
    expected_warn = (
        "Land surface fraction (sftlf) data not available, using default instead"
    )
    with patch.dict(
        WEIGHTS_FUNCTIONS_WITHOUT_AREA_WEIGHTING,
        {"World|Northern Hemisphere": mock_nh_weights},
    ):
        if not with_bounds:
            test_all_cubes.lat_dim.bounds = None
            test_all_cubes.lon_dim.bounds = None
        weighter = CubeWeightCalculator(test_all_cubes)
        result = weighter.get_weights(expected.keys())

    if not with_bounds:
        assert len(caplog.messages) == 4
        assert (
            caplog.messages[0]
            == "Couldn't find/use areacell_cube, falling back to iris.analysis.cartography.area_weights"
        )
        assert caplog.messages[0] == caplog.messages[1]
        assert caplog.messages[2] == "Guessing latitude and longitude bounds"
        assert caplog.messages[3] == expected_warn
    else:
        assert len(caplog.messages) == 3
        assert (
            caplog.messages[0]
            == "Couldn't find/use areacell_cube, falling back to iris.analysis.cartography.area_weights"
        )
        assert caplog.messages[0] == caplog.messages[1]
        assert caplog.messages[2] == expected_warn

    for label, array in expected.items():
        np.testing.assert_array_equal(array, result[label])
    mock_nh_weights.assert_called_with(weighter, test_all_cubes)


@pytest.mark.parametrize("input_format", ["scmcube", None])
@pytest.mark.parametrize("sftlf_var", ["sftlf", "sftlf_other"])
@patch(
    "netcdf_scm.iris_cube_wrappers.SCMCube.surface_fraction_var",
    new_callable=PropertyMock,
)
def test_get_land_weights(
    mock_surface_fraction_var, test_all_cubes, input_format, sftlf_var
):
    sftlf_cube = create_sftlf_cube(test_all_cubes.__class__)
    mock_surface_fraction_var.return_value = sftlf_var
    original_data = sftlf_cube.cube.data

    test_all_cubes.get_metadata_cube = MagicMock(return_value=sftlf_cube)

    test_land_fraction_input = sftlf_cube if input_format == "scmcube" else None

    weighter = CubeWeightCalculator(test_all_cubes)
    result = get_land_weights(
        weighter, test_all_cubes, sftlf_cube=test_land_fraction_input
    )

    expected = original_data
    np.testing.assert_array_equal(result, expected)
    # Check that the sftlf meta cube is always registered
    test_all_cubes.get_metadata_cube.assert_called_with(
        sftlf_var, cube=test_land_fraction_input
    )


@pytest.mark.parametrize("surface_frac_var", ["sftlf", "sftof"])
@patch(
    "netcdf_scm.iris_cube_wrappers.SCMCube.netcdf_scm_realm", new_callable=PropertyMock
)
def test_get_land_weights_shape_errors(
    mock_netcdf_scm_realm, surface_frac_var, test_all_cubes
):
    surface_frac_cube = create_sftlf_cube(test_all_cubes.__class__)
    error_msg = re.escape(
        "the {} cube data must be the same shape as the "
        "cube's longitude-latitude grid".format(surface_frac_var)
    )

    wrong_shape_data = np.array([[1, 2], [3, 4]])
    surface_frac_cube.cube = iris.cube.Cube(data=wrong_shape_data)

    mock_netcdf_scm_realm.return_value = (
        "land" if surface_frac_var == "sftlf" else "ocean"
    )

    weighter = CubeWeightCalculator(test_all_cubes)
    with pytest.raises(AssertionError, match=error_msg):
        if surface_frac_var == "sftlf":
            get_land_weights(weighter, test_all_cubes, sftlf_cube=surface_frac_cube)
        else:
            get_ocean_weights(weighter, test_all_cubes, sftof_cube=surface_frac_cube)

    test_all_cubes.get_metadata_cube = MagicMock(return_value=surface_frac_cube)
    with pytest.raises(AssertionError, match=error_msg):
        if surface_frac_var == "sftlf":
            get_land_weights(weighter, test_all_cubes, sftlf_cube=None)
        else:
            get_ocean_weights(weighter, test_all_cubes, sftof_cube=None)


def create_dummy_cube_from_lat_lon_points(lat_pts, lon_pts):
    lat = iris.coords.DimCoord(lat_pts, standard_name="latitude", units="degrees")
    lon = iris.coords.DimCoord(
        lon_pts, standard_name="longitude", units="degrees", circular=True
    )

    cube = iris.cube.Cube(
        np.full((len(lat_pts), len(lon_pts)), 0),
        dim_coords_and_dims=[(lat, 0), (lon, 1)],
    )

    return cube


def test_nao_weights(test_all_cubes):
    sftlf_cube = create_sftlf_cube(test_all_cubes.__class__)
    weighter = CubeWeightCalculator(test_all_cubes, sftlf_cube=sftlf_cube)
    result = weighter.get_weights_array_without_area_weighting(
        "World|North Atlantic Ocean"
    )

    expected = np.array([[0, 0, 0, 1], [0, 0, 0, 1], [0, 0, 0, 0]]) * (
        100 - sftlf_cube.cube.data
    )

    np.testing.assert_array_equal(result, expected)


def test_elnino_weights(test_all_cubes):
    sftlf_cube = create_sftlf_cube(test_all_cubes.__class__)
    weighter = CubeWeightCalculator(test_all_cubes, sftlf_cube=sftlf_cube)
    result = weighter.get_weights_array_without_area_weighting("World|El Nino N3.4")
    # 5N-5S, 190E-240E
    expected = np.array([[0, 0, 0, 0], [0, 0, 1, 0], [0, 0, 0, 0]]) * (
        100 - sftlf_cube.cube.data
    )

    np.testing.assert_array_equal(result, expected)


@pytest.mark.parametrize(
    "lat_pts,lon_pts,expected",
    [
        (  # nothing within bounds, raises Error
            np.array([-60, -1, 80]),
            np.array([45, 135, 225, 315]),
            "error",
        ),
        (  # nothing within bounds, raises Error
            np.array([-60, 10, 80]),
            np.array([45, 135, 225, 279]),
            "error",
        ),
        (  # nothing within bounds negative co-ord, raises Error
            np.array([-60, -1, 80]),
            np.array([-135, -45, 45, 135]),
            "error",
        ),
        (  # edge of bound included
            np.array([65, 0, -60]),
            np.array([45, 135, 225, 280]),
            np.array([[0, 0, 0, 1], [0, 0, 0, 1], [0, 0, 0, 0]]),
        ),
        (  # edge of bound included negative co-ord
            np.array([66, 0, -60]),
            np.array([-135, -80, 45, 135]),
            np.array([[0, 0, 0, 0], [0, 1, 0, 0], [0, 0, 0, 0]]),
        ),
        (  # one within bounds
            np.array([80, 35, -70]),
            np.array([10, 30, 50, 135, 320]),
            np.array([[0, 0, 0, 0, 0], [0, 0, 0, 0, 1], [0, 0, 0, 0, 0]]),
        ),
        (  # one within bounds negative co-ord
            np.array([80, 35, -70]),
            np.array([-95, -40, 40, 135]),
            np.array([[0, 0, 0, 0], [0, 1, 0, 0], [0, 0, 0, 0]]),
        ),
    ],
)
@pytest.mark.parametrize("query", [[0, -80, 65, 0], [0, 280, 65, 360]])
def test_get_weights_for_area(test_all_cubes, query, lat_pts, lon_pts, expected):
    regrid_cube = create_dummy_cube_from_lat_lon_points(lat_pts, lon_pts)
    test_all_cubes.cube = test_all_cubes.cube.regrid(
        regrid_cube, iris.analysis.Linear()
    )

    if isinstance(expected, str) and expected == "error":
        error_msg = re.compile("None of the cube's.*lie within the bounds.*")
        with pytest.raises(ValueError, match=error_msg):
            get_weights_for_area(*query)(None, test_all_cubes)
        return

    result = get_weights_for_area(*query)(None, test_all_cubes)
    np.testing.assert_array_equal(result, expected)


def test_area_mask_wrapped_lons(test_all_cubes):
    result = get_weights_for_area(0, -80, 65, 0)(None, test_all_cubes)

    expected = np.array([[0, 0, 0, 1], [0, 0, 0, 1], [0, 0, 0, 0]])

    np.testing.assert_array_equal(result, expected)


def test_get_masks_unknown_weights_warning(test_all_cubes, caplog):
    caplog.set_level(logging.WARNING, logger="netcdf_scm.iris_cube_wrappers")
    weighter = CubeWeightCalculator(test_all_cubes)
    res = weighter.get_weights(["World", "junk"])

    np.testing.assert_allclose(res["World"], test_all_cubes.get_area_weights())

    assert len(caplog.messages) == 3
    assert (
        caplog.messages[0]
        == "Couldn't find/use areacell_cube, falling back to iris.analysis.cartography.area_weights"
    )
    assert caplog.records[0].levelname == "WARNING"
    assert caplog.messages[1] == "Failed to create junk weights: Unknown weights: junk"
    assert caplog.records[1].levelname == "WARNING"
    assert caplog.messages[0] == caplog.messages[2]
    assert caplog.records[0].levelname == caplog.records[2].levelname


@pytest.mark.parametrize("exp_warn,cube_max", [(False, 100), (True, 1)])
@pytest.mark.parametrize("surface_frac_var", ["sftlf", "sftof"])
@patch(
    "netcdf_scm.iris_cube_wrappers.SCMCube.surface_fraction_var",
    new_callable=PropertyMock,
)
def test_get_scm_weights_surface_frac_bound_checks(
    mock_surface_fraction_var,
    exp_warn,
    cube_max,
    surface_frac_var,
    test_all_cubes,
    caplog,
):
    tsurface_frac_cube = get_default_sftlf_cube().regrid(
        test_all_cubes.cube, iris.analysis.Linear()
    )
    tsftlf_cube_max = tsurface_frac_cube.data.max()
    assert np.isclose(tsftlf_cube_max, 100)
    if cube_max == 1:
        tsurface_frac_cube.data = tsurface_frac_cube.data / 100

    mock_surface_fraction_var.return_value = surface_frac_var
    test_all_cubes.get_metadata_cube = MagicMock()
    tsurface_frac_scmcube = SCMCube()
    tsurface_frac_scmcube.cube = tsurface_frac_cube

    tothermdata_scmcube = SCMCube()
    tothermdata_scmcube.cube = tsurface_frac_cube.copy()
    tothermdata_scmcube.cube.units = "m**2"

    def mock_meta_getter(metadata_variable, cube):
        if metadata_variable == test_all_cubes.surface_fraction_var:
            return tsurface_frac_scmcube

        return tothermdata_scmcube

    test_all_cubes.get_metadata_cube.side_effect = mock_meta_getter

    caplog.set_level(logging.INFO)
    weighter = CubeWeightCalculator(test_all_cubes)
    weighter.get_weights(
        ["World|Land"] if surface_frac_var == "sftlf" else ["World|Ocean"]
    )

    test_all_cubes.get_metadata_cube.assert_has_calls(
        [call(test_all_cubes.surface_fraction_var, cube=None)]
    )

    if exp_warn:
        expected_warn = "{} data max is {}, multiplying by 100 to convert units to percent".format(
            surface_frac_var, tsurface_frac_cube.data.max()
        )
        assert len(caplog.messages) == 2
        assert caplog.messages[1] == expected_warn
    else:
        assert len(caplog.messages) == 1

    assert "NetCDF-SCM will treat the data as `atmosphere`" in caplog.messages[0]



"""Input and output from NetCDF-SCM's netCDF format"""
try:
    import iris
except ModuleNotFoundError:  # pragma: no cover # emergency valve
    from .errors import raise_no_iris_warning

    raise_no_iris_warning()

from .definitions import _SCM_TIMESERIES_META_COLUMNS
from .iris_cube_wrappers import SCMCube


def save_netcdf_scm_nc(cubes, out_path):
    """
    Save a series of cubes to a `.nc` file

    Parameters
    ----------
    cubes : dict
        Dictionary of "region name"-:obj:`ScmCube` key-value pairs. The cubes will all
        be saved in the same ``.nc`` file.

    out_path : str
        Path in which to save the data
    """
    save_cubes = []
    for scm_cube in cubes.values():
        cube = scm_cube.cube
        save_cubes.append(cube)

    iris.save(save_cubes, out_path, local_keys=_SCM_TIMESERIES_META_COLUMNS)


def load_scmdataframe(path):
    """
    Load an scmdataframe from a NetCDF-SCM ``.nc`` file

    Parameters
    ----------
    path : str
        Path from which to load the data

    Returns
    -------
    :obj:`ScmDataFrame`
        :obj:`ScmDataFrame` containing the data in ``path``.
    """
    helper, scm_cubes = _load_helper_and_scm_cubes(path)
    scmdf = helper.convert_scm_timeseries_cubes_to_openscmdata(scm_cubes)

    # TODO: decide whether assuming point data is a good idea or not
    return scmdf


def _load_helper_and_scm_cubes(path):
    cube_list = iris.load(path)

    loaded = SCMCube()
    scm_cubes = {}
    for v in cube_list:
        region = v.attributes["region"]
        scm_cubes[region] = SCMCube()
        scm_cubes[region].cube = v

    # take any cube as base for now, not sure how to really handle this so will
    # leave like this for now and only make this method public when I work it
    # out...
    loaded.cube = cube_list[0]

    return loaded, scm_cubes



import versioneer
from setuptools import find_packages, setup
from setuptools.command.test import test as TestCommand

PACKAGE_NAME = "netcdf-scm"
DESCRIPTION = "Processing netCDF files for use with simple climate models"
KEYWORDS = [
    "netcdf",
    "netCDF",
    "python",
    "climate",
    "atmosphere",
    "simple climate model",
    "reduced complexity climate model",
    "data processing",
]

AUTHOR = "Zebedee Nicholls"
EMAIL = "zebedee.nicholls@climate-energy-college.org"
URL = "https://github.com/znicholls/netcdf-scm"
PROJECT_URLS = {
    "Bug Reports": "https://github.com/znicholls/netcdf-scm/issues",
    "Documentation": "https://openscm.readthedocs.io/en/latest",
    "Source": "https://github.com/znicholls/netcdf-scm",
}
LICENSE = "3-Clause BSD License"
CLASSIFIERS = [
    "Development Status :: 4 - Beta",
    "License :: OSI Approved :: BSD License",
    "Intended Audience :: Developers",
    "Operating System :: OS Independent",
    "Programming Language :: Python :: 3.7",
]

ENTRY_POINTS = {
    "console_scripts": [
        "netcdf-scm-crunch = netcdf_scm.cli:crunch_data",
        "netcdf-scm-wrangle = netcdf_scm.cli:wrangle_netcdf_scm_ncs",
        "netcdf-scm-stitch = netcdf_scm.cli:stitch_netcdf_scm_ncs",
    ]
}


REQUIREMENTS_INSTALL = [
    "numpy",
    "python-dateutil",
    "progressbar2",
    "pymagicc>=2.0.0b8",
    "click",
    "tqdm",
]
REQUIREMENTS_TESTS = [
    "codecov",
    "pytest-cov",
    "pytest>=4.0,<5.0",
    "pytest-console-scripts",
]
REQUIREMENTS_NOTEBOOKS = [
    "notebook",
    "nbval",
    "expectexception",
    "pyam-iamc>=0.2.0",
    "pandas==0.24.2",
]
REQUIREMENTS_DOCS = ["sphinx>=1.4,<2.1", "sphinx_rtd_theme", "sphinx-click"]
REQUIREMENTS_DEPLOY = ["twine>=1.11.0", "setuptools>=38.6.0", "wheel>=0.31.0"]
requirements_dev = [
    *[
        "flake8",
        "black",
        "isort",
        "bandit",
        "mypy",
        "pydocstyle",
        "pylint>=2.4.0",
        "nbdime",
    ],
    *REQUIREMENTS_TESTS,
    *REQUIREMENTS_NOTEBOOKS,
    *REQUIREMENTS_DOCS,
    *REQUIREMENTS_DEPLOY,
]

REQUIREMENTS_EXTRAS = {
    "docs": REQUIREMENTS_DOCS,
    "tests": REQUIREMENTS_TESTS,
    "notebooks": REQUIREMENTS_NOTEBOOKS,
    "deploy": REQUIREMENTS_DEPLOY,
    "dev": requirements_dev,
}


SOURCE_DIR = "src"

# no tests/docs in `src` so don't need exclude
PACKAGES = find_packages(SOURCE_DIR)
PACKAGE_DIR = {"": SOURCE_DIR}
PACKAGE_DATA = {"netcdf_scm": ["weights/*.nc"]}


README = "README.rst"

with open(README, "r") as readme_file:
    README_TEXT = readme_file.read()


class NetCDFSCM(TestCommand):
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        import pytest

        pytest.main(self.test_args)


cmdclass = versioneer.get_cmdclass()
cmdclass.update({"test": NetCDFSCM})

setup(
    name=PACKAGE_NAME,
    version=versioneer.get_version(),
    description=DESCRIPTION,
    long_description=README_TEXT,
    long_description_content_type="text/x-rst",
    author=AUTHOR,
    author_email=EMAIL,
    url=URL,
    project_urls=PROJECT_URLS,
    license=LICENSE,
    classifiers=CLASSIFIERS,
    keywords=KEYWORDS,
    packages=PACKAGES,
    package_dir=PACKAGE_DIR,
    package_data=PACKAGE_DATA,
    include_package_data=True,
    install_requires=REQUIREMENTS_INSTALL,
    extras_require=REQUIREMENTS_EXTRAS,
    cmdclass=cmdclass,
    entry_points=ENTRY_POINTS,
)



"""
Utils contains a number of helpful functions for doing common cube operations.

For example, applying masks to cubes, taking latitude-longitude means and getting
timeseries from a cube as datetime values.
"""
import datetime as dt
import logging

import numpy as np
import numpy.ma as ma
from dateutil.relativedelta import relativedelta

try:
    import cftime
    import dask.array as da
    import iris
    from iris.analysis import WeightedAggregator, _build_dask_mdtol_function
    from iris.exceptions import CoordinateNotFoundError
    import cf_units

    # monkey patch iris MEAN until https://github.com/SciTools/iris/pull/3299 is merged
    iris.analysis.MEAN = WeightedAggregator(
        "mean", ma.average, lazy_func=_build_dask_mdtol_function(da.ma.average)
    )

except ModuleNotFoundError:  # pragma: no cover # emergency valve
    from .errors import raise_no_iris_warning

    raise_no_iris_warning()

logger = logging.getLogger(__name__)


def get_cube_timeseries_data(scm_cube, realise_data=False):
    """
    Get a timeseries from a cube.

    This function only works on cubes which are on a time grid only i.e. have no other dimension coordinates.

    Parameters
    ----------
    scm_cube : :obj:`SCMCube`
        An ``SCMCube`` instance with only a 'time' dimension.

    realise_data : bool
        If ``True``, force the data to be realised before returning

    Returns
    -------
    np.ndarray
        The cube's timeseries data. If ``realise_data`` is ``False`` then a
        ``da.Array`` will be returned if the data is lazy.
    """
    _assert_only_cube_dim_coord_is_time(scm_cube)
    raw_data = scm_cube.cube.core_data()
    if isinstance(raw_data, da.Array) and realise_data:
        return raw_data.compute()
    return raw_data


def get_scm_cube_time_axis_in_calendar(scm_cube, calendar):
    """
    Get a cube's time axis in a given calendar

    Parameters
    ----------
    scm_cube : :obj:`SCMCube`
        An ``SCMCube`` instance.
    calendar : str
        The calendar to return the time axis in e.g. '365_day', 'gregorian'.

    Returns
    -------
    np.ndarray
        Array of datetimes, containing the cube's calendar.
    """
    time = scm_cube.cube.dim_coords[scm_cube.time_dim_number]
    return cf_units.num2date(time.points, time.units.name, calendar)


def _assert_only_cube_dim_coord_is_time(scm_cube):
    assert_msg = "Should only have time coordinate here"
    if len(scm_cube.cube.dim_coords) != 1:
        raise AssertionError(assert_msg)
    if scm_cube.cube.dim_coords[0].standard_name != "time":
        raise AssertionError(assert_msg)


def assert_all_time_axes_same(time_axes):
    """
    Assert all time axes in a set are the same.

    Parameters
    ----------
    time_axes : list_like of array_like
        List of time axes to compare.

    Raises
    ------
    AssertionError
        If not all time axes are the same.
    """
    for time_axis_to_check in time_axes:
        assert_msg = "all the time axes should be the same"
        try:
            np.testing.assert_array_equal(
                time_axis_to_check, time_axes[0], err_msg=assert_msg
            )
        # handle weird numpy error in case it comes up
        except AttributeError:  # pragma: no cover
            raise AssertionError(assert_msg)


def take_lat_lon_mean(in_scmcube, in_weights):
    """
    Take the latitude longitude mean of a cube with given weights

    Parameters
    ----------
    in_scmcube : :obj:`SCMCube`
        An ``SCMCube`` instance.

    in_weights : np.ndarray
        Weights to use when taking the mean.

    Returns
    -------
    :obj:`SCMCube`
        A copy of the input cube in which the data is now the latitude-longitude mean
        of the input cube's data
    """
    out_cube = type(in_scmcube)()
    if in_weights.shape != in_scmcube.cube.shape:
        in_weights = broadcast_onto_lat_lon_grid(in_scmcube, in_weights)

    out_cube.cube = in_scmcube.cube.collapsed(
        [in_scmcube.lat_name, in_scmcube.lon_name],
        iris.analysis.MEAN,
        weights=in_weights,
    )
    return out_cube


def apply_mask(in_scmcube, in_mask):
    """
    Apply a mask to an scm cube's data

    Parameters
    ----------
    in_scmcube : :obj:`SCMCube`
        An ``SCMCube`` instance.

    in_mask : np.ndarray
        The mask to apply

    Returns
    -------
    :obj:`SCMCube`
        A copy of the input cube with the mask applied to its data
    """
    out_cube = type(in_scmcube)()
    if in_scmcube.cube.has_lazy_data():
        new_data = da.ma.masked_array(data=in_scmcube.cube.lazy_data(), mask=in_mask)
    else:
        new_data = ma.masked_array(in_scmcube.cube.data, mask=in_mask)
    out_cube.cube = in_scmcube.cube.copy(data=new_data)

    return out_cube


def unify_lat_lon(cubes, rtol=10 ** -6):
    """
    Unify latitude and longitude co-ordinates of cubes in place.

    The co-ordinates will only be unified if they already match to within a given
    tolerance.

    Parameters
    ----------
    cubes : :obj:`iris.cube.CubeList`
        List of iris cubes whose latitude and longitude co-ordinates should be unified.

    rtol : float
        Maximum relative difference which can be accepted between co-ordinate values.

    Raises
    ------
    ValueError
        If the co-ordinates differ by more than relative tolerance or are not
        compatible (e.g. different shape).
    """
    ref_lats = cubes[0].coords("latitude")[0].points
    ref_lons = cubes[0].coords("longitude")[0].points
    for cube in cubes[1:]:
        cube_lats = cube.coords("latitude")[0].points
        cube_lons = cube.coords("longitude")[0].points
        try:
            np.testing.assert_allclose(ref_lats, cube_lats, rtol=rtol)
            np.testing.assert_allclose(ref_lons, cube_lons, rtol=rtol)
        except AssertionError:
            error_msg = (
                "Cannot unify latitude and longitude, relative difference in "
                "co-ordinates is greater than {}".format(rtol)
            )
            raise ValueError(error_msg)

        lat_dim_numbers = cube.coord_dims("latitude")
        cube.remove_coord("latitude")
        if len(lat_dim_numbers) > 1:
            cube.add_aux_coord(cubes[0].coords("latitude")[0], lat_dim_numbers)
        else:
            cube.add_dim_coord(cubes[0].coords("latitude")[0], lat_dim_numbers[0])

        lon_dim_numbers = cube.coord_dims("longitude")
        cube.remove_coord("longitude")
        if len(lon_dim_numbers) > 1:
            cube.add_aux_coord(cubes[0].coords("longitude")[0], lon_dim_numbers)
        else:
            cube.add_dim_coord(cubes[0].coords("longitude")[0], lon_dim_numbers[0])


def cube_lat_lon_grid_compatible_with_array(cube, array_in):
    """
    Assert that an array can be broadcast onto the cube's lat-lon grid

    Parameters
    ----------
    cube : :obj:`ScmCube`
        :obj:`ScmCube` instance whose lat-lon grid we want to check agains


    array_in : np.ndarray
        The array we want to ensure is able to be broadcast

    Returns
    -------
    bool
        ``True`` if the cube's lat-lon grid is compatible with ``array_in``, otherwise
        ``False``

    Raises
    ------
    AssertionError
        The array cannot be broadcast onto the cube's lat-lon grid
    """
    base_shape = cube.lat_lon_shape

    if array_in.shape != base_shape:
        array_in = np.transpose(array_in)

    if array_in.shape != base_shape:
        return False

    return True


def broadcast_onto_lat_lon_grid(cube, array_in):
    """
    Broadcast an array onto the latitude-longitude grid of ``cube``.

    Here, broadcasting means taking the array and 'duplicating' it so that it
    has the same number of dimensions as the cube's underlying data.

    For example, given a cube with a time dimension of length 3, a latitude dimension of length 4
    and a longitude dimension of length 2 (shape 3x4x2) and ``array_in`` of shape 4x2, results in
    a 3x4x2 array where each slice in the broadcasted array's time dimension is identical to ``array_in``.

    Parameters
    ----------
    cube : :obj:`ScmCube`
        :obj:`ScmCube` instance whose lat-lon grid we want to check agains

    array_in : np.ndarray
        The array we want to broadcast

    Returns
    -------
    array_out
        The original array, broadcast onto the cube's lat-lon grid (i.e. duplicated
        along all dimensions except for latitude and longitude). Note: If the cube has
        lazy data, we return a :obj:`da.Array`, otherwise we return an
        :obj:`np.ndarray`.

    Raises
    ------
    AssertionError
        ``array_in`` cannot be broadcast onto the cube's lat-lon grid because their
        shapes are not compatible

    ValueError
        ``array_in`` cannot be broadcast onto the cube's lat-lon grid by
        ``iris.util.broadcast_to_shape``
    """
    if not cube_lat_lon_grid_compatible_with_array(cube, array_in):
        shape_assert_msg = (
            "the ``array_in`` must be the same shape as the "
            "cube's longitude-latitude grid"
        )
        raise AssertionError(shape_assert_msg)

    broadcaster = np.broadcast_to if not cube.cube.has_lazy_data() else da.broadcast_to
    try:
        return broadcaster(array_in, cube.cube.shape)
    except ValueError as e:
        try_transpose = str(e).startswith(
            "operands could not be broadcast together with remapped shapes"
        ) or str(e).startswith("cannot broadcast shape")
        if not try_transpose:  # pragma: no cover
            raise
        return broadcaster(array_in.T, cube.cube.shape)


def _cftime_conversion(t):
    return dt.datetime(
        t.year, t.month, t.day, t.hour, t.minute, t.second, t.microsecond
    )


_vector_cftime_conversion = np.vectorize(_cftime_conversion)


def _check_cube_and_adjust_if_needed(cube, time_name="time"):
    """
    Check cube and adjust if required

    Parameters
    ----------
    cube : :obj:`iris.cube.Cube`
        Cube to check

    time_name : str
        Name of the time dimension of the cube to check

    Returns
    -------
    :obj:`iris.cube.Cube`
        Cube, adjusted if needed
    """
    try:
        time_dim = cube.coord(time_name)
        gregorian = time_dim.units.calendar == "gregorian"
        year_zero = str(time_dim.units).startswith("days since 0-1-1")
    except CoordinateNotFoundError:
        gregorian = False
        year_zero = False

    if gregorian and year_zero:
        warn_msg = (
            "Your calendar is gregorian yet has units of 'days since 0-1-1'. "
            "We rectify this by removing all data before year 1 and changing the "
            "units to 'days since 1-1-1'. If you want other behaviour, you will "
            "need to use another package."
        )
        logger.warning(warn_msg)
        return _adjust_gregorian_year_zero_units(cube, time_name)

    return cube


def _adjust_gregorian_year_zero_units(cube, time_name):
    """
    Adjust Gregogrian calendar with year zero.

    This function makes the time axis useable with iris again (there is no year zero
    in the Gregorian calendar) at the expense of removing the year zero data.

    Parameters
    ----------
    cube : :obj:`iris.cube.Cube`
        Cube to adjusted

    time_name : str
        Name of the time dimension of the cube to adjust

    Returns
    -------
    :obj:`iris.cube.Cube`
        Adjusted cube

    Raises
    ------
    AssertionError
        Defensive assertion: the code is being used in an unexpected way
    """
    # pylint:disable=too-many-locals
    # hack function to work around very specific use case
    year_zero_cube = cube.copy()
    year_zero_cube_time_dim = cube.coord(time_name)

    gregorian_year_zero_cube = (
        year_zero_cube_time_dim.units.calendar == "gregorian"
    ) and str(year_zero_cube_time_dim.units).startswith("days since 0-1-1")
    if not gregorian_year_zero_cube:  # pragma: no cover # emergency valve
        raise AssertionError("This function is not setup for other cases")

    new_unit_str = "days since 1-1-1"
    # converting with the new units means we're actually converting with the wrong
    # units, we use this variable to track how many years to shift back to get the
    # right time axis again
    new_units_shift = 1
    new_time_dim_unit = cf_units.Unit(
        new_unit_str, calendar=year_zero_cube_time_dim.units.calendar
    )

    tmp_time_dim = year_zero_cube_time_dim.copy()
    tmp_time_dim.units = new_time_dim_unit
    tmp_cube = iris.cube.Cube(year_zero_cube.data)
    for i, coord in enumerate(year_zero_cube.dim_coords):
        if coord.standard_name == "time":
            tmp_cube.add_dim_coord(tmp_time_dim, i)
        else:
            tmp_cube.add_dim_coord(coord, i)

    years_to_bin = 1
    first_valid_year = years_to_bin + new_units_shift

    def check_usable_data(cell):
        return first_valid_year <= cell.point.year

    usable_cube = tmp_cube.extract(iris.Constraint(time=check_usable_data))
    usable_data = usable_cube.data

    tmp_time_dim = usable_cube.coord(time_name)
    tmp_time = cftime.num2date(
        tmp_time_dim.points, new_unit_str, tmp_time_dim.units.calendar
    )
    # TODO: move to utils
    tmp_time = np.array([dt.datetime(*v.timetuple()[:6]) for v in tmp_time])
    # undo the shift to new units
    usable_time = cf_units.date2num(
        tmp_time - relativedelta(years=new_units_shift),
        year_zero_cube_time_dim.units.name,
        year_zero_cube_time_dim.units.calendar,
    )
    usable_time_unit = cf_units.Unit(
        year_zero_cube_time_dim.units.name,
        calendar=year_zero_cube_time_dim.units.calendar,
    )
    usable_time_dim = iris.coords.DimCoord(
        usable_time,
        standard_name=year_zero_cube_time_dim.standard_name,
        long_name=year_zero_cube_time_dim.long_name,
        var_name=year_zero_cube_time_dim.var_name,
        units=usable_time_unit,
    )

    new_cube = iris.cube.Cube(usable_data)
    for i, coord in enumerate(usable_cube.dim_coords):
        if coord.standard_name == "time":
            new_cube.add_dim_coord(usable_time_dim, i)
        else:
            new_cube.add_dim_coord(coord, i)

    # hard coding as making this list dynamically is super hard as there's so many
    # edge cases to cover
    attributes_to_copy = [
        "attributes",
        "cell_methods",
        "units",
        "var_name",
        "standard_name",
        "name",
        "metadata",
        "long_name",
    ]
    for att in attributes_to_copy:
        setattr(new_cube, att, getattr(year_zero_cube, att))

    return new_cube
